/*
 * Created on November 9, 2005, 3:37 PM
 *
 */

package com.sunlabs.archsim.fleet.ships.selector;

import com.sunlabs.archsim.base.hierarchy.CompositeEntity;
import com.sunlabs.archsim.base.simulation.Command;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.Payload;
import com.sunlabs.archsim.gasp.Action;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.fleet.common.OutputFunction;
import com.sunlabs.archsim.gasp.Function;
import com.sunlabs.archsim.gasp.State;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;
import com.sunlabs.archsim.globals.Globals;

// import java.util.Iterator;

/**
 *
 * @author mh193489
 */
public class SelectorAction extends Action {
    private Delay adderDelay= SelectorDefaults.SELECTOR_DELAY_DEF;
    
    Function outFunction= new DupplicateData();
    Function ackInFunction= new AckFunction();
    
    /** references to terminals attached to this action */
    InputTerminal in1;
    OutputTerminal ackIn1;
    InputTerminal in2;
    OutputTerminal ackIn2;
    InputTerminal inSelect;
    OutputTerminal ackInSelect;
    
    OutputTerminal out;
    InputTerminal ackOut;
    
    /** counts the number of pending duplicates to be published */
//    private int NUM_DUPLICATES= 2;
//    private int count= NUM_DUPLICATES;
    
    /** delay reset untiul the next data on In arrives */
//    private boolean resetOnNextIn= true;
    
    private int branch;
    
    /**
     * mark active states with booleans, to prevent multiple invocations of
     * fireBranch() when several states become active aimultaneousely
     */
    private boolean activeInSelect, activeAckOut, activeIn1, activeIn2;
    
    
    private boolean consumeIn1, consumeIn2;
    
    private Data data;
    
    public SelectorAction(String name, CompositeEntity parent) {
        super(name, parent);
        outputFunctions.put(SelectorDefaults.OUT, outFunction);
        outputFunctions.put(SelectorDefaults.ACK_IN1, ackInFunction);
        outputFunctions.put(SelectorDefaults.ACK_IN2, ackInFunction);
        outputFunctions.put(SelectorDefaults.ACK_IN_SELECT, ackInFunction);
        
        notifyStateIsActiveCommand= new StateActiveCommand();
        notifyStateIsInactiveCommand= new StateInactiveCommand();
    } // SelectorAction
    
    public void setDelay(Delay d) {
        adderDelay= (Delay)d.clone();
        outFunction.setDelay(adderDelay);
        ackInFunction.setDelay(adderDelay);
    }
    
    public boolean selfCheck() {
        boolean OK= true;
        boolean check;
        
        check= (this.getTerminal(SelectorDefaults.IN1) != null);
        OK= OK && check;
        if (!check) {
            director.error("Action " + getAlias() + ", id= " + getID()
            + ", does not have attached terminal in1");
        } else {
            in1= (InputTerminal)this.getTerminal(SelectorDefaults.IN1);
        }
        
        check= (this.getTerminal(SelectorDefaults.ACK_IN1) != null);
        OK= OK && check;
        if (!check) {
            director.error("Action " + getAlias() + ", id= " + getID()
            + ", does not have attached terminal ackIn1");
        } else {
            ackIn1= (OutputTerminal)this.getTerminal(SelectorDefaults.ACK_IN1);
        }
        
        check= (this.getTerminal(SelectorDefaults.IN2) != null);
        OK= OK && check;
        if (!check) {
            director.error("Action " + getAlias() + ", id= " + getID()
            + ", does not have attached terminal in2");
        } else {
            in2= (InputTerminal)this.getTerminal(SelectorDefaults.IN2);
        }
        
        check= (this.getTerminal(SelectorDefaults.ACK_IN2) != null);
        OK= OK && check;
        if (!check) {
            director.error("Action " + getAlias() + ", id= " + getID()
            + ", does not have attached terminal ackIn2");
        } else {
            ackIn2= (OutputTerminal)this.getTerminal(SelectorDefaults.ACK_IN2);
        }
        
        check= (this.getTerminal(SelectorDefaults.IN_SELECT) != null);
        OK= OK && check;
        if (!check) {
            director.error("Action " + getAlias() + ", id= " + getID()
            + ", does not have attached terminal inSelect");
        } else {
            inSelect= (InputTerminal)this.getTerminal(SelectorDefaults.IN_SELECT);
        }
        
        check= (this.getTerminal(SelectorDefaults.ACK_IN_SELECT) != null);
        OK= OK && check;
        if (!check) {
            director.error("Action " + getAlias() + ", id= " + getID()
            + ", does not have attached terminal ackInSelect");
        } else {
            ackInSelect= (OutputTerminal)this.getTerminal(SelectorDefaults.ACK_IN_SELECT);
        }
        
        
        check= (this.getTerminal(SelectorDefaults.OUT) != null);
        OK= OK && check;
        if (!check) {
            director.error("Action " + getAlias() + ", id= " + getID()
            + ", does not have attached terminal out");
        } else {
            out= (OutputTerminal)this.getTerminal(SelectorDefaults.OUT);
        }
        
        check= (this.getTerminal(SelectorDefaults.ACK_OUT) != null);
        OK= OK && check;
        if (!check) {
            director.error("Action " + getAlias() + ", id= " + getID()
            + ", does not have attached terminal ackOut");
        } else {
            ackOut= (InputTerminal)this.getTerminal(SelectorDefaults.ACK_OUT);
        }
        
        
        // also run the self check on the superclass
        check= super.selfCheck();
        OK= OK && check;
        
        return OK;
    }
    
    private void fireBranch() {
        if (activeInSelect && (branch == -1)) {
            inSelect.setStateInactive();
            computeAndFire(ackInSelect);
            activeInSelect= false;
            
            consumeIn1= true;
            consumeIn2= true;
        }
        
        if ((branch == -1) && activeIn1) {
//            director.error("SELECTOR consuming in1");
            data= ((Payload)in1.getData()).data;
            if (consumeIn1) {
                // if OOB, make this the last consumption
                if (!data.valid && (data.value == null)) {
                    consumeIn1= false;
                }
                in1.setStateInactive();
                computeAndFire(ackIn1);
                activeIn1= false;
            }
            if (!consumeIn1 && !consumeIn2) {
                branch= -2;
            }
        }
        
        if ((branch == -1) && activeIn2) {
//            director.error("SELECTOR consuming in2");
            data= ((Payload)in2.getData()).data;
            if (consumeIn2) {
                // if OOB, make this the last consumption
                if (!data.valid && (data.value == null)) {
                    consumeIn2= false;
                }
                in2.setStateInactive();
                computeAndFire(ackIn2);
                activeIn2= false;
            }
            if (!consumeIn1 && !consumeIn2) {
                branch= -2;
            }
        }
        
        if (activeInSelect && (branch == 1) && activeAckOut && activeIn1 /* && in2.getState().isActive() */) {
//            director.error("SELECTOR fired from in1");
            data= ((Payload)in1.getData()).data;
            ackOut.setStateInactive();
            inSelect.setStateInactive();
            in1.setStateInactive();
            computeAndFire(out);
            computeAndFire(ackInSelect);
            computeAndFire(ackIn1);
            
            activeInSelect= false;
            activeIn1= false;
            
        } else if (activeInSelect && (branch == 2) && activeAckOut && activeIn2 /* && in1.getState().isActive() */) {
//            director.error("SELECTOR fired from in2");
            data= ((Payload)in2.getData()).data;
            ackOut.setStateInactive();
            inSelect.setStateInactive();
            in2.setStateInactive();
            computeAndFire(out);
            computeAndFire(ackInSelect);
            computeAndFire(ackIn2);
            
            activeInSelect= false;
            activeIn2= false;
            
        }
/*
        else {
            director.error("SELECTOR not fired because ...");
            if (!activeInSelect) {
                director.error("  activeInSelect) == false");
            }
            if (!activeAckOut) {
                director.error("  activeAckOut == false");
            }
            if (!(branch == 1) && !(branch == 2)) {
                director.error("  !(branch == 1) && !(branch == 2)");
            }
            if (!activeIn1) {
                director.error("  in1 not active");
            }
            if (!activeIn2) {
                director.error("  in2 not active");
            }
        }
 */
    }
    
    protected class StateActiveCommand extends Command {
        public void execute(Object param) {
            // update the count of active inputs.
            // note that the state itself contains a check
            // whether it became active twice without being
            // inactive in between
            Terminal src= (Terminal)param;
            if (journalAction) {
                State state= src.getState();
                director.journal("Action " + getAlias()
                + ", id= " + getID() + " notifed that state "
                        + state.getAlias() + ", id= " + state.getID()
                        + " is active");
            }
            
            if (src == in1) {                
                activeIn1= true;
                fireBranch();
                                
            } else if (src == in2) {                
                activeIn2= true;
                fireBranch();
                                
            } else if (src == inSelect) {
                // check if finished last LAST
                if (!(consumeIn1 || consumeIn2)) {
                    activeInSelect= true;
                    // get branch from data on inSelect
                    if (((Payload)inSelect.getData()).data != null) {
                        if (((Payload)inSelect.getData()).data.value != null) {
                            branch= (Integer)((Payload)inSelect.getData()).data.value;
                        } else {
                            if (!((Payload)inSelect.getData()).data.valid) {
                                branch= -1;
                            } else {
//                                director.error("SELECTOR Received data.value == null and data.valid = true on inSelect");
                            }
                            
                        }
                        
                    } else {
//                        director.error("SELECTOR Received data == null on inSelect");
                        
                    }
                }
                fireBranch();
                
                
            } else if (src == ackOut) {                
                activeAckOut= true;
                fireBranch();
                
            }
            
        } // execute
    } // StateActiveCommand
    
    
    protected class StateInactiveCommand extends Command {
        public void execute(Object param) {
            Terminal src= (Terminal)param;
            
            if (src == in1) {
                activeIn1= false;
                
            } else if (src == in2) {
                activeIn2= false;
                
            } else if (src == inSelect) {
                activeInSelect= false;
                
            } else if (src == ackOut) {
                activeAckOut= false;
            }
            
            if (journalAction) {
                State state= src.getState();
                director.journal("Action " + getAlias()
                + ", id= " + getID() + " notifed that state "
                        + state.getAlias() + ", id= " + state.getID()
                        + " is INactive");
            }
        } // execute
    } // StateInactiveCommand
    
    
    private class DupplicateData extends Function {
        
        public DupplicateData() {
            super(SelectorDefaults.ADD_DATA);
            functionDelay= adderDelay;
        }
        
        public Object value() {
            return data;
        }
        
    } // value
    
    
    private class AckFunction extends OutputFunction {
        
        /** Global settings */
        protected Globals globals= Globals.getInstance();
        
        public AckFunction() {
            this("AckFunction");
        }
        
        public AckFunction(String name) {
            super(name);
            Integer fDelay= globals.intValue(SelectorDefaults.SELCTOR_ACK_DELAY);
            functionDelay= (fDelay != null)
            ? new Delay(fDelay.intValue())
            : SelectorDefaults.SELECTOR_ACK_DELAY_DEF;
        }
        
        
        public Data value() {
            return new Data(true);
        } // value
        
    } // AckFunction
}


